/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
const scriptMgr = Cc["@thumbsdown.mozdev.org/script-manager;1"].getService().wrappedJSObject;


const TOPIC_TARGET_HEAD = "thumbsdown-scriptwizard-target-head";

const TARGET_IMAGE = 1;
const TARGET_PAGE  = 2;

const MODE_CREATE = 1;
const MODE_MODIFY = 2;

const TYPE_DIRECT  = 1;
const TYPE_REGEXP  = 2;
const TYPE_CUSTOM  = 3;
const TYPE_WRAPPED = 4;

const TYPE_REGEXP_THUMBNAIL = 1;
const TYPE_REGEXP_LINK      = 2;

const FILTER_MODE_BOTH   = 1;  // both width and height over limit
const FILTER_MODE_EITHER = 2;  // either width or height over limit

const INPUT_DELAY = 100;


// Most of the 'depends' here are unnecassary because of pages' dirty flags.
// Afaict, only 'matchingScript', 'linkUrlTokens' and 'target' are really needed (see ThumbnailsPage.onPageAdvanced())
var modelConfig =
{
    referer: {},
    
    thumbnails: {                       value: [], equals: arrayEquals},
    selected:   {depends: "thumbnails", value: [], equals: arrayEquals},
    images:     {depends: "selected",   value: [], equals: arrayEquals},
    
    matchingScript: {depends: "selected"},
    linkUrlTokens:  {depends: "selected"},
    target:         {depends: "selected", value: 0},
    
    mode:   {depends: "selected", value: 0},
    script: {depends: "selected mode", value: null, equals: function(a, b) { return ((a === b) || ((a !== null) && (b !== null) && (a.id === b.id))); }},
    
    name:        {depends: "selected script"},
    matcher:     {depends: "selected script"},
    type:        {depends: "selected target script", value: 0},
    source:      {depends: "images script", value: 0},
    pattern:     {depends: "images script"},
    template:    {depends: "images script"},
    attrName:    {depends: "images script"},
    attrValue:   {depends: "images script"},
    
    error: {}
}


var scriptWizard =
{
    init: function()
    {
        stringBundle = $("stringbundle");
        
        var prefButton = wizard.elem.getButton("extra1");
        prefButton.label = _("preferences_");
        prefButton.hidden = false;
        prefButton.addEventListener("command", function() { scriptWizard.openPreferences(); }, false);
        
        wizard.elem.setAttribute("onwizardfinish", "return scriptWizard.onWizardFinish()");
        
        model = new Model(modelConfig);
        
        var doc = window.arguments[0];
        model.referer = doc.location.href;
        
        var thumbnails = [];
        var images = doc.images;
        if (images)
        {
            loop:
            for (var i = 0; i < images.length; i++)
            {
                var thumbnail = Thumbnail.new(images[i]);
                if (! thumbnail)
                    continue loop;
                
                for (var k = 0; k < thumbnails.length; k++)
                    if (thumbnail.equals(thumbnails[k]))
                        continue loop;
                
                thumbnail.index = thumbnails.length;
                thumbnails.push(thumbnail);
            }
        }
        model.thumbnails = thumbnails;
        
        // let ui draw before initThumbnailTargets()
        callLater(0, this, this.initThumbnailTargets);
    },
    
    initThumbnailTargets: function()
    {
        var observerSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        var timeout = nsPreferences.getIntPref("extensions.thumbsdown.scriptwizard.requestTimeout") * 1000;
        var thumbnails = model.thumbnails;
        
        for (var i = 0; i < thumbnails.length; i++)
        {
            let thumbnail = thumbnails[i];
            
            var callback = function(loader)
            {
                var target = thumbnail.target.head;
                
                if (loader.success)
                {
                    target.result = "unknown";
                    target.status = loader.request.status;
                    target.contentType = loader.request.getResponseHeader("Content-Type");
                    if (target.contentType)
                    {
                        var match =  /^([-\w]+\/[-\w]+)/.exec(target.contentType);
                        if (match)
                        {
                            target.mimeType = match[1];
                            if (/^image\/\w+/.test(target.mimeType))
                                target.result = "image";
                            else if (target.mimeType == "text/html")
                                target.result = "page";
                        }
                    }
                }
                else
                {
                    target.result = (loader.timeout ? "timeout" : "error");
                    if (! loader.timeout)
                        target.status = loader.request.status;
                }
                
                observerSvc.notifyObservers(thumbnail, TOPIC_TARGET_HEAD, thumbnail.index);
            };
            
            thumbnail.target.head = {};
            thumbnail.target.head.loader = new HeadLoader(thumbnail.anchor.link, model.referer, timeout, callback);
        }
    },
    
    getMatchingScript: function(thumbnails)
    {
        // look up all the existing scripts that match for each thumbnail
        var matchesArray = [];
        for each (var thumbnail in thumbnails)
        {
            var matches = scriptMgr.lookupAll(thumbnail.anchor.link, thumbnail.image.link);
            
            // it's not possible to say that wrapped script matches solely based on a thumbnail (attribute must match also)
            // of course, the same is true with custom script also, but we'll ignore that for now
            matches = [script for each (script in matches) if (script.type != TYPE_WRAPPED)]
            
            if (matches.length == 0)
                break;
            
            matchesArray.push(matches);
        }
        
        // is there an existing script that matches all the thumbnails
        if (matchesArray.length == thumbnails.length)
        {
            loop:
            for (var i = 0; i < matchesArray[0].length; i++)
            {
                var script = matchesArray[0][i];
                var isSameScript = function(s) { return (s.id == script.id); };
                for (var k = 1; k < matchesArray.length; k++)
                    if (! matchesArray[k].some(isSameScript))
                        continue loop;
                
                return script;
            }
        }
        
        return null;
    },
    
    getLinkUrlTokens: function(thumbnails)
    {
        var urls = [thumbnail.anchor.url for each (thumbnail in thumbnails)];
        var urlTokens = new UrlTokens(urls);
        
        if (! urlTokens.schemes.equal)
            throw "scheme.notEqual";
        
        if (! urlTokens.hostTokens.slice(1).every(function(sa) { return sa.equal; }))
            throw "host.notEqual";
        
        if (urlTokens.pathTokens[urlTokens.pathTokens.length - 1].hasNull)
            throw "path.notEqualCount";
        
        return urlTokens;
    },
    
    getName: function(urlTokens)
    {
        var hostToken = (urlTokens.hostTokens[0].equal ? urlTokens.hostTokens[0].strings[0] : null);
        var index = ((hostToken && (! /^www/.test(hostToken))) ? 0 : 1);
        var name = [sa.strings[0] for each (sa in urlTokens.hostTokens.slice(index))].join(".");
        return name;
    },
    
    getMatcher: function(urlTokens)
    {
        return this.resolveUrlRegExp(urlTokens);
    },
    
    getPattern: function(name, thumbnails)
    {
        var urls = [thumbnail[name].url for each (thumbnail in thumbnails)];
        var urlTokens = new UrlTokens(urls);
        return this.resolveUrlRegExp(urlTokens, true);
    },
    
    getTemplate: function(name, thumbnails, pattern)
    {
        var targets = new StringArray([thumbnail.target.image.link for each (thumbnail in thumbnails)]);
        
        // add common prefix to template and remove it from targets
        var template = targets.prefix;
        targets = targets.substring(targets.prefix.length);
        if (targets.isEmpty)
            return template;
        
        var sources = [thumbnail[name].link for each (thumbnail in thumbnails)];
        var re = new RegExp("^" + pattern + "$");
        
        // collect match object for each source
        var matchesArray = [];
        for (var i = 0; i < thumbnails.length; i++)
            matchesArray[i] = re.exec(sources[i]);
        
        // bail out if pattern did not match all the sources
        if (matchesArray.some(function(value) { return (! value); }))
            return template + targets.strings[0];
        
        // collect captures into StringArrays
        var saMatches = [];
        for (var i = 1; i < matchesArray[0].length; i++)
            saMatches.push(new StringArray([match[i] for each (match in matchesArray)]));
        
        // replace "dynamic" parts in targets with captures from sources
        // TODO: captures are used in order. while usually true, not necessarily
        for (var index = 0; index < saMatches.length; index++)
        {
            var matches = saMatches[index];
            if (targets.startsWith(matches))
            {
                template += "$" + (index + 1);
                targets = targets.substr(matches.lengths());
                template += targets.prefix;
                targets = targets.substr(targets.prefix.length);
            }
        }
        
        if (! targets.isEmpty)
            return template + targets.strings[0];
        
        return template;
    },
    
    getWrappedAttributeRegExp: function(thumbnails, attrName)
    {
        var values = [];
        for each (var thumbnail in thumbnails)
            if (thumbnail.target.image[attrName])
                values.push(thumbnail.target.image[attrName]);
            else
                return null;
        
        return this.resolveRegExp(new StringArray(values));
    },
    
    getWrappedUrlAttributeRegExp: function(thumbnails, attrName)
    {
        var urls = [thumbnail.target.image[attrName] for each (thumbnail in thumbnails)];
        var urlTokens = new UrlTokens(urls);
        return this.resolveUrlRegExp(urlTokens, false);
    },
    
    resolveUrlRegExp: function(urlTokens, capture)
    {
        // TODO: eh, ever heard about non-greedy quantifiers? replace sentinels?
        var sentinels;
        var tokens;
        
        var hasQueries = (! urlTokens.queries.isEmpty);
        var hasRefs = (! urlTokens.refs.isEmpty);
        
        // scheme
        var scheme = this.resolveRegExp(urlTokens.schemes);
        
        // host
        sentinels = arrayWithValue(urlTokens.hostTokens.length, ".");
        sentinels[sentinels.length - 1] = "/";
        tokens = [];
        for (var i = 0; i < urlTokens.hostTokens.length; i++)
            tokens.push(this.resolveRegExp(urlTokens.hostTokens[i], capture, sentinels[i]));
        var host = tokens.join(escapeRegExp("."));
        
        // path
        sentinels = arrayWithValue(urlTokens.pathTokens.length, "/");
        sentinels[sentinels.length - 1] = (hasQueries ? "?" : (hasRefs ? "#" : null))
        tokens = [];
        for (var i = 0; i < urlTokens.pathTokens.length; i++)
            tokens.push(this.resolveRegExp(urlTokens.pathTokens[i], capture, sentinels[i]));
        var path = tokens.join(escapeRegExp("/"));
        
        // query
        sentinels = arrayWithValue(urlTokens.queryTokens.length, "&");
        sentinels[sentinels.length - 1] = (hasRefs ? "#" : null);
        tokens = [];
        for (var i = 0; i < urlTokens.queryTokens.length; i++)
            tokens.push(this.resolveRegExp(urlTokens.queryTokens[i], capture, sentinels[i]));
        var query = tokens.join(escapeRegExp("&"));
        if (query)
            query = escapeRegExp("?") + query;
        
        // reference
        var reference = this.resolveRegExp(urlTokens.refs);
        if (reference)
            reference = escapeRegExp("#") + reference;
        
        return scheme + "://" + host + path + query + reference;
    },
    
    resolveRegExp: function(sa, capture, sentinel)
    {
        if (sa.equal)
            return escapeRegExp(sa.strings[0]);
        
        var prefix = escapeRegExp(sa.prefix);
        var middle = this._resolveRegExp(sa.middle(), sentinel);
        var suffix = escapeRegExp(sa.suffix);
        
        if (capture)
            middle = "(" + middle +")";
        
        return prefix + middle + suffix;
    },
    
    _resolveRegExp: function(sa, sentinel)
    {
        var p = "";
        var n = sa.size;
        var metaCharClass = null;
        var charClass = "";
        var pattern = null;
        var quantifier = ((sa.length.min == 0) ? "*" : "+");
        
        var i = 0, k = 0;
        while ((i < n) && (sa.strings[i].length == 0))
            i++;
        
        while (i < n)
        {
            var str = sa.strings[i];
            var c = str.charAt(k);
            
            if ((! metaCharClass) && /\d/.test(c))
                metaCharClass = "\\d";
            else if ((metaCharClass != "\\w") && /\w/.test(c))
                metaCharClass = "\\w";
            else if (charClass.length < 1)
                charClass = ((c == "-") ? c + charClass : charClass + c);
            else if (! pattern)
                pattern = (sentinel ? "[^" + sentinel + "]" : ".");
            else
                throw new Error("unreachable code");
            
            var p = (pattern ? pattern : (charClass ? "[" + escapeRegExp(charClass) + metaCharClass + "]" : metaCharClass)) + quantifier;
            var re = new RegExp("^" + p);
            
            for (i = 0; i < n; i++)
            {
                str = sa.strings[i];
                var match = re.exec(str);
                if (! match)
                {
                    k = 0;
                    break;
                }
                if (match[0].length != str.length)
                {
                    k = match[0].length;
                    break;
                }
            }
        }
        
        return p;
    },
    
    openPreferences: function()
    {
        // TODO: use openPrefWindow from util
        var windowUrl = "chrome://thumbsdown/content/scriptwizard/prefwindow.xul";
        var windowName = "thumbsdown_scriptwizard_prefwindow";
        var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
        var win = windowMediator.getMostRecentWindow(windowName);
        if (win)
            win.focus();
        else
        {
            var instantApply = nsPreferences.getBoolPref("browser.preferences.instantApply", false);
            var features = "chrome,titlebar,toolbar,centerscreen," + (instantApply ? "dialog=no" : "dialog=yes");
            features = features + ",centerscreen=no,dependent";
            window.openDialog(windowUrl, windowName, features);
        }
    },
    
    onWizardFinish: function()
    {
        if (model.matchingScript)
            return true;
        
        var script = {};
        
        if (model.mode == MODE_MODIFY)
            script.id = model.script.id;
        
        script.name = model.name;
        script.matcher = model.matcher;
        script.type = model.type;
        
        switch (model.type)
        {
            case TYPE_REGEXP:
                script.subType = model.source;
                script.action1 = model.pattern;
                script.action2 = model.template;
                break;
                
            case TYPE_WRAPPED:
                script.action1 = model.attrName;
                script.action2 = model.attrValue;
                break;
        }
        
        if (model.mode == MODE_MODIFY)
            scriptMgr.setScript(script);
        else
            scriptMgr.addScript(script);
        
        return true;
    }
}

wizard.addLoadListener(function() { scriptWizard.init(); });


function Thumbnail(image)
{
    this.image = {};
    this.image.elem = image;
    this.image.link = image.src;
    this.image.url = ioSvc.newURI(image.src, null, null).QueryInterface(Ci.nsIURL);
    
    var anchor = image.parentNode;
    this.anchor = {};
    this.anchor.elem = anchor;
    this.anchor.link = anchor.href;
    this.anchor.url = ioSvc.newURI(anchor.href, null, null).QueryInterface(Ci.nsIURL);
    
    this.target = {
        /*
        head: {
            loader:       // HeadLoader
            result:       // 
            status:       // http status
            contentType:  // http content type header
            mimeType:     // mime-type extracted from content type
        }
        loader:     // ResourceLoader to load the target page
        images: {   // list of images from the target page
            id:     // img.id
            alt:    // img.alt
            link:   // img.src
            url:    // img.src as nsiURL
            index:  // index in array
        },
        image:      // selected image. reference to images[]
        */
    };
}

Thumbnail.new = function(image)
{
    try
    {
        return new Thumbnail(image);
    }
    catch (e)
    {
        return null;
    }
}

Thumbnail.prototype =
{
    equals: function(thumbnail)
    {
        return ((this.image.link == thumbnail.image.link) && (this.anchor.link == thumbnail.anchor.link));
    }
}


function UrlTokens(urls)
{
    this.urls = urls;
    
    this.schemes = new StringArray([url.scheme   for each (url in urls)]);
    this.hosts   = new StringArray([url.host     for each (url in urls)]);
    this.paths   = new StringArray([url.filePath for each (url in urls)]);
    this.queries = new StringArray([url.query    for each (url in urls)]);
    this.refs    = new StringArray([url.ref      for each (url in urls)]);
    
    this.hostTokens = this.hosts.tokenize(".");
    this.pathTokens = this.paths.tokenize("/");
    this.queryTokens = this.queries.tokenize("&");
}


function HeadLoader(link, referer, timeout, callback)
{
    this.callback = callback;
    
    // TODO: use nsiXHR? XHR "stops" when wizard page is changed
    this.request = new XMLHttpRequest();
    this.request.open("HEAD", link, true);
    this.request.setRequestHeader("Referer", referer);
    var self = this;
    this.request.onload = function() { self.wrap(self.onLoad); };
    this.request.onerror = function() { self.wrap(self.onError); };
    this.request.send(null);
    
    this.timer = callLater(timeout, this, this.wrap, this.onTimeout);
}

HeadLoader.prototype =
{
    wrap: function(func)
    {
        try
        {
            if (this.success === undefined)
                func.apply(this);
            if (this.callback)
                this.callback(this);
        }
        finally
        {
            this.destroy();
        }
    },
    
    onLoad: function()
    {
        this.success = true;
    },
    
    onError: function()
    {
        this.success = false;
    },
    
    onTimeout: function()
    {
        this.success = false;
        this.timeout = true;
    },
    
    destroy: function()
    {
        if (this.request)
        {
            this.request.abort();
            delete this.request;
        }
        
        if (this.timer)
        {
            this.timer.cancel();
            delete this.timer;
        }
        
        delete this.callback;
    }
}
